@@ -16,75 +16,34 @@ nevertheless, and we (the maintainers) do make some efforts to not needlessly
 change the public API.
 """
 
+# TODO: When Python 3.9 is dropped, consider using `type | None` instead of
+# `Optional[type]`.
+
 import gettext
 import logging
 import os
-import re
 from dataclasses import dataclass, field
 from enum import Enum
 from importlib.metadata import PackageNotFoundError, version
-from typing import Any, Dict, NamedTuple, Optional, Set, Type
+from typing import Any, Optional
 
 from boolean.boolean import Expression
+from license_expression import Licensing
 
 try:
     __version__ = version("reuse")
 except PackageNotFoundError:
     # package is not installed
-    __version__ = "4.0.3"
+    __version__ = "5.0.2"
 
 __author__ = "Carmen Bianca Bakker"
 __email__ = "carmenbianca@fsfe.org"
 __license__ = "Apache-2.0 AND CC0-1.0 AND CC-BY-SA-4.0 AND GPL-3.0-or-later"
-__REUSE_version__ = "3.2"
+__REUSE_version__ = "3.3"
 
 _LOGGER = logging.getLogger(__name__)
 
-_PACKAGE_PATH = os.path.dirname(__file__)
-_LOCALE_DIR = os.path.join(_PACKAGE_PATH, "locale")
-
-if gettext.find("reuse", localedir=_LOCALE_DIR):
-    gettext.bindtextdomain("reuse", _LOCALE_DIR)
-    gettext.textdomain("reuse")
-    _LOGGER.debug("translations found at %s", _LOCALE_DIR)
-else:
-    _LOGGER.debug("no translations found at %s", _LOCALE_DIR)
-
-
-_IGNORE_DIR_PATTERNS = [
-    re.compile(r"^\.git$"),
-    re.compile(r"^\.hg$"),
-    re.compile(r"^\.sl$"),  # Used by Sapling SCM
-    re.compile(r"^LICENSES$"),
-    re.compile(r"^\.reuse$"),
-]
-
-_IGNORE_MESON_PARENT_DIR_PATTERNS = [
-    re.compile(r"^subprojects$"),
-]
-
-_IGNORE_FILE_PATTERNS = [
-    re.compile(r"^LICENSE(\..*)?$"),
-    re.compile(r"^COPYING(\..*)?$"),
-    # ".git" as file happens in submodules
-    re.compile(r"^\.git$"),
-    re.compile(r"^\.gitkeep$"),
-    re.compile(r"^\.hgtags$"),
-    re.compile(r".*\.license$"),
-    # Workaround for https://github.com/fsfe/reuse-tool/issues/229
-    re.compile(r"^CAL-1.0(-Combined-Work-Exception)?(\..+)?$"),
-    re.compile(r"^SHL-2.1(\..+)?$"),
-]
-
-_IGNORE_SPDX_PATTERNS = [
-    # SPDX files from
-    # https://spdx.github.io/spdx-spec/conformance/#44-standard-data-format-requirements
-    re.compile(r".*\.spdx$"),
-    re.compile(r".*\.spdx.(rdf|json|xml|ya?ml)$"),
-]
-
-# Combine SPDX patterns into file patterns to ease default ignore usage
-_IGNORE_FILE_PATTERNS.extend(_IGNORE_SPDX_PATTERNS)
+_LICENSING = Licensing()
 
 
 class SourceType(Enum):
@@ -107,9 +66,9 @@ class SourceType(Enum):
 class ReuseInfo:
     """Simple dataclass holding licensing and copyright information"""
 
-    spdx_expressions: Set[Expression] = field(default_factory=set)
-    copyright_lines: Set[str] = field(default_factory=set)
-    contributor_lines: Set[str] = field(default_factory=set)
+    spdx_expressions: set[Expression] = field(default_factory=set)
+    copyright_lines: set[str] = field(default_factory=set)
+    contributor_lines: set[str] = field(default_factory=set)
     path: Optional[str] = None
     source_path: Optional[str] = None
     source_type: Optional[SourceType] = None
@@ -133,10 +92,10 @@ class ReuseInfo:
         return self.__class__(**new_kwargs)  # type: ignore
 
     def union(self, value: "ReuseInfo") -> "ReuseInfo":
-        """Return a new instance of ReuseInfo where all Set attributes are equal
+        """Return a new instance of ReuseInfo where all set attributes are equal
         to the union of the set in *self* and the set in *value*.
 
-        All non-Set attributes are set to their values in *self*.
+        All non-set attributes are set to their values in *self*.
 
         >>> one = ReuseInfo(copyright_lines={"Jane Doe"}, source_path="foo.py")
         >>> two = ReuseInfo(copyright_lines={"John Doe"}, source_path="bar.py")
@@ -178,11 +137,3 @@ class ReuseInfo:
 
     def __or__(self, value: "ReuseInfo") -> "ReuseInfo":
         return self.union(value)
-
-
-class ReuseException(Exception):
-    """Base exception."""
-
-
-class IdentifierNotFound(ReuseException):
-    """Could not find SPDX identifier for license file."""
