@@ -1,27 +1,47 @@
 # SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
 # SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
 # SPDX-FileCopyrightText: 2022 Pietro Albini <pietro.albini@ferrous-systems.com>
+# SPDX-FileCopyrightText: 2023 DB Systel GmbH
 # SPDX-FileCopyrightText: 2023 Carmen Bianca BAKKER <carmenbianca@fsfe.org>
+# SPDX-FileCopyrightText: 2024 Kerry McAdams <github@klmcadams>
+# SPDX-FileCopyrightText: 2024 Sebastien Morais <github@SMoraisAnsys>
 #
 # SPDX-License-Identifier: GPL-3.0-or-later
 
 """Module that contains reports about files and projects for linting."""
 
+import bdb
+import contextlib
 import datetime
 import logging
 import multiprocessing as mp
 import random
-from gettext import gettext as _
 from hashlib import md5
 from io import StringIO
 from os import cpu_count
 from pathlib import Path, PurePath
-from typing import Any, Dict, Iterable, List, NamedTuple, Optional, Set, cast
+from typing import (
+    Any,
+    Collection,
+    Iterable,
+    NamedTuple,
+    Optional,
+    Protocol,
+    cast,
+)
 from uuid import uuid4
 
-from . import __REUSE_version__, __version__
-from ._util import _LICENSING, StrPath, _checksum
+from . import _LICENSING, __REUSE_version__, __version__
+from ._util import (
+    _add_plus_to_identifier,
+    _checksum,
+    _strip_plus_from_identifier,
+)
+from .extract import _LICENSEREF_PATTERN
+from .global_licensing import ReuseDep5
+from .i18n import _
 from .project import Project, ReuseInfo
+from .types import StrPath
 
 _LOGGER = logging.getLogger(__name__)
 
@@ -34,11 +54,44 @@ class _MultiprocessingContainer:
     def __init__(
         self, project: Project, do_checksum: bool, add_license_concluded: bool
     ):
-        self.project = project
+        if isinstance(project.global_licensing, ReuseDep5):
+            # Remember that a dep5_copyright was (or was not) set prior.
+            self.has_dep5 = bool(project.global_licensing)
+            # TODO: We create a copy of the project in the following
+            # song-and-dance because the debian Copyright object cannot be
+            # pickled.
+            new_project = Project(
+                project.root,
+                vcs_strategy=project.vcs_strategy,
+                license_map=project.license_map,
+                licenses=project.licenses.copy(),
+                # TODO: adjust this method/class to account for REUSE.toml as
+                # well. Unset dep5_copyright
+                global_licensing=None,
+                include_submodules=project.include_submodules,
+                include_meson_subprojects=project.include_meson_subprojects,
+            )
+            new_project.licenses_without_extension = (
+                project.licenses_without_extension
+            )
+            self.project = new_project
+        else:
+            self.has_dep5 = False
+            self.project = project
+
+        self.reuse_dep5: Optional[ReuseDep5] = None
         self.do_checksum = do_checksum
         self.add_license_concluded = add_license_concluded
 
     def __call__(self, file_: StrPath) -> "_MultiprocessingResult":
+        # By remembering that we've parsed the .reuse/dep5, we only parse it
+        # once (the first time) inside of each process.
+        if self.has_dep5 and not self.reuse_dep5:
+            with contextlib.suppress(Exception):
+                self.reuse_dep5 = ReuseDep5.from_file(
+                    self.project.root / ".reuse/dep5"
+                )
+                self.project.global_licensing = self.reuse_dep5
         # pylint: disable=broad-except
         try:
             return _MultiprocessingResult(
@@ -63,35 +116,107 @@ class _MultiprocessingResult(NamedTuple):
     error: Optional[Exception]
 
 
+def _generate_file_reports(
+    project: Project,
+    do_checksum: bool = True,
+    subset_files: Optional[Collection[StrPath]] = None,
+    multiprocessing: bool = cpu_count() > 1,  # type: ignore
+    add_license_concluded: bool = False,
+) -> Iterable[_MultiprocessingResult]:
+    """Create a :class:`FileReport` for every file in the project, filtered
+    by *subset_files*.
+    """
+    container = _MultiprocessingContainer(
+        project, do_checksum, add_license_concluded
+    )
+
+    files = (
+        project.subset_files(subset_files)
+        if subset_files is not None
+        else project.all_files()
+    )
+    if multiprocessing:
+        with mp.Pool() as pool:
+            results: Iterable[_MultiprocessingResult] = pool.map(
+                container, files
+            )
+        pool.join()
+    else:
+        results = map(container, files)
+    return results
+
+
+def _process_error(error: Exception, path: StrPath) -> None:
+    # Facilitate better debugging by being able to quit the program.
+    if isinstance(error, (bdb.BdbQuit, KeyboardInterrupt)):
+        raise error
+    if isinstance(error, (OSError, UnicodeError)):
+        _LOGGER.error(
+            _("Could not read '{path}'").format(path=path),
+            exc_info=error,
+        )
+    else:
+        _LOGGER.error(
+            _("Unexpected error occurred while parsing '{path}'").format(
+                path=path
+            ),
+            exc_info=error,
+        )
+
+
+class ProjectReportSubsetProtocol(Protocol):
+    """A :class:`Protocol` that defines a subset of functionality of
+    :class:`ProjectReport`, implemented by :class:`ProjectSubsetReport`.
+    """
+
+    path: StrPath
+    missing_licenses: dict[str, set[Path]]
+    read_errors: set[Path]
+    file_reports: set["FileReport"]
+
+    @property
+    def files_without_licenses(self) -> set[Path]:
+        """Set of paths that have no licensing information."""
+
+    @property
+    def files_without_copyright(self) -> set[Path]:
+        """Set of paths that have no copyright information."""
+
+    @property
+    def is_compliant(self) -> bool:
+        """Whether the report subset is compliant with the REUSE Spec."""
+
+
 class ProjectReport:  # pylint: disable=too-many-instance-attributes
     """Object that holds linting report about the project."""
 
     def __init__(self, do_checksum: bool = True):
         self.path: StrPath = ""
-        self.licenses: Dict[str, Path] = {}
-        self.missing_licenses: Dict[str, Set[Path]] = {}
-        self.bad_licenses: Dict[str, Set[Path]] = {}
-        self.deprecated_licenses: Set[str] = set()
-        self.read_errors: Set[Path] = set()
-        self.file_reports: Set[FileReport] = set()
-        self.licenses_without_extension: Dict[str, Path] = {}
+        self.licenses: dict[str, Path] = {}
+        self.missing_licenses: dict[str, set[Path]] = {}
+        self.bad_licenses: dict[str, set[Path]] = {}
+        self.deprecated_licenses: set[str] = set()
+        self.read_errors: set[Path] = set()
+        self.file_reports: set[FileReport] = set()
+        self.licenses_without_extension: dict[str, Path] = {}
 
         self.do_checksum = do_checksum
 
-        self._unused_licenses: Optional[Set[str]] = None
-        self._used_licenses: Optional[Set[str]] = None
-        self._files_without_licenses: Optional[Set[Path]] = None
-        self._files_without_copyright: Optional[Set[Path]] = None
+        self._unused_licenses: Optional[set[str]] = None
+        self._used_licenses: Optional[set[str]] = None
+        self._files_without_licenses: Optional[set[Path]] = None
+        self._files_without_copyright: Optional[set[Path]] = None
         self._is_compliant: Optional[bool] = None
 
-    def to_dict_lint(self) -> Dict[str, Any]:
+    def to_dict_lint(self) -> dict[str, Any]:
         """Collects and formats data relevant to linting from report and returns
         it as a dictionary.
 
-        :return: Dictionary containing data from the ProjectReport object
+        Returns:
+            Dictionary containing data from the ProjectReport object.
         """
         # Setup report data container
-        data: Dict[str, Any] = {
+        data: dict[str, Any] = {
             "non_compliant": {
                 "missing_licenses": self.missing_licenses,
                 "unused_licenses": [str(file) for file in self.unused_licenses],
@@ -112,6 +237,7 @@ class ProjectReport:  # pylint: disable=too-many-instance-attributes
             "summary": {
                 "used_licenses": [],
             },
+            "recommendations": self.recommendations,
         }
 
         # Populate 'files'
@@ -139,15 +265,21 @@ class ProjectReport:  # pylint: disable=too-many-instance-attributes
         }
 
         # Sort dictionary keys while keeping the top three keys at the beginning
+        # and the recommendations on the bottom
         sorted_keys = sorted(list(unsorted_data.keys()))
         sorted_keys.remove("lint_version")
         sorted_keys.remove("reuse_spec_version")
         sorted_keys.remove("reuse_tool_version")
-        sorted_keys = [
-            "lint_version",
-            "reuse_spec_version",
-            "reuse_tool_version",
-        ] + sorted_keys
+        sorted_keys.remove("recommendations")
+        sorted_keys = (
+            [
+                "lint_version",
+                "reuse_spec_version",
+                "reuse_tool_version",
+            ]
+            + sorted_keys
+            + ["recommendations"]
+        )
 
         sorted_data = {key: unsorted_data[key] for key in sorted_keys}
 
@@ -182,9 +314,8 @@ class ProjectReport:  # pylint: disable=too-many-instance-attributes
         )
         out.write(f"Creator: Tool: reuse-{__version__}\n")
 
-        now = datetime.datetime.utcnow()
-        now = now.replace(microsecond=0)
-        out.write(f"Created: {now.isoformat()}Z\n")
+        now = datetime.datetime.now(tz=datetime.timezone.utc)
+        out.write(f"Created: {now.strftime('%Y-%m-%dT%H:%M:%SZ')}\n")
         out.write(
             "CreatorComment: <text>This document was created automatically"
             " using available reuse information consistent with"
@@ -195,7 +326,7 @@ class ProjectReport:  # pylint: disable=too-many-instance-attributes
 
         for report in reports:
             out.write(
-                "Relationship: SPDXRef-DOCUMENT describes"
+                "Relationship: SPDXRef-DOCUMENT DESCRIBES"
                 f" {report.spdx_id}\n"
             )
 
@@ -217,7 +348,7 @@ class ProjectReport:  # pylint: disable=too-many-instance-attributes
 
         # Licenses
         for lic, path in sorted(self.licenses.items()):
-            if lic.startswith("LicenseRef-"):
+            if _LICENSEREF_PATTERN.match(lic):
                 out.write("\n")
                 out.write(f"LicenseID: {lic}\n")
                 out.write("LicenseName: NOASSERTION\n")
@@ -235,7 +366,16 @@ class ProjectReport:  # pylint: disable=too-many-instance-attributes
         multiprocessing: bool = cpu_count() > 1,  # type: ignore
         add_license_concluded: bool = False,
     ) -> "ProjectReport":
-        """Generate a ProjectReport from a Project."""
+        """Generate a :class:`ProjectReport` from a :class:`Project`.
+
+        Args:
+            project: The :class:`Project` to lint.
+            do_checksum: Generate a checksum of every file. If this is
+                :const:`False`, generate a random checksum for every file.
+            multiprocessing: Whether to use multiprocessing.
+            add_license_concluded: Whether to aggregate all found SPDX
+                expressions into a concluded license.
+        """
         project_report = cls(do_checksum=do_checksum)
         project_report.path = project.root
         project_report.licenses = project.licenses
@@ -243,36 +383,18 @@ class ProjectReport:  # pylint: disable=too-many-instance-attributes
             project.licenses_without_extension
         )
 
-        container = _MultiprocessingContainer(
-            project, do_checksum, add_license_concluded
+        results = _generate_file_reports(
+            project,
+            do_checksum=do_checksum,
+            multiprocessing=multiprocessing,
+            add_license_concluded=add_license_concluded,
         )
-
-        if multiprocessing:
-            with mp.Pool() as pool:
-                results: Iterable[_MultiprocessingResult] = pool.map(
-                    container, project.all_files()
-                )
-            pool.join()
-        else:
-            results = map(container, project.all_files())
-
         for result in results:
             if result.error:
-                if isinstance(result.error, (OSError, UnicodeError)):
-                    _LOGGER.error(
-                        _("Could not read '{path}'").format(path=result.path),
-                        exc_info=result.error,
-                    )
-                    project_report.read_errors.add(Path(result.path))
-                    continue
-                _LOGGER.error(
-                    _(
-                        "Unexpected error occurred while parsing '{path}'"
-                    ).format(path=result.path),
-                    exc_info=result.error,
-                )
+                _process_error(result.error, result.path)
                 project_report.read_errors.add(Path(result.path))
                 continue
+
             file_report = cast(FileReport, result.report)
 
             # File report.
@@ -300,7 +422,7 @@ class ProjectReport:  # pylint: disable=too-many-instance-attributes
         return project_report
 
     @property
-    def used_licenses(self) -> Set[str]:
+    def used_licenses(self) -> set[str]:
         """Set of license identifiers that are found in file reports."""
         if self._used_licenses is not None:
             return self._used_licenses
@@ -313,26 +435,24 @@ class ProjectReport:  # pylint: disable=too-many-instance-attributes
         return self._used_licenses
 
     @property
-    def unused_licenses(self) -> Set[str]:
+    def unused_licenses(self) -> set[str]:
         """Set of license identifiers that are not found in any file report."""
         if self._unused_licenses is not None:
             return self._unused_licenses
 
-        # First collect licenses that are suspected to be unused.
-        suspected_unused_licenses = {
-            lic for lic in self.licenses if lic not in self.used_licenses
-        }
-        # Remove false positives.
         self._unused_licenses = {
             lic
-            for lic in suspected_unused_licenses
-            if f"{lic}+" not in self.used_licenses
+            for lic in self.licenses
+            if not any(
+                identifier in self.used_licenses
+                for identifier in set((lic, _add_plus_to_identifier(lic)))
+            )
         }
         return self._unused_licenses
 
     @property
-    def files_without_licenses(self) -> Set[Path]:
-        """Set of paths that have no license information."""
+    def files_without_licenses(self) -> set[Path]:
+        """Set of paths that have no licensing information."""
         if self._files_without_licenses is not None:
             return self._files_without_licenses
 
@@ -345,7 +465,7 @@ class ProjectReport:  # pylint: disable=too-many-instance-attributes
         return self._files_without_licenses
 
     @property
-    def files_without_copyright(self) -> Set[Path]:
+    def files_without_copyright(self) -> set[Path]:
         """Set of paths that have no copyright information."""
         if self._files_without_copyright is not None:
             return self._files_without_copyright
@@ -379,6 +499,169 @@ class ProjectReport:  # pylint: disable=too-many-instance-attributes
 
         return self._is_compliant
 
+    @property
+    def recommendations(self) -> list[str]:
+        """Generate help for next steps based on found REUSE issues"""
+        recommendations = []
+
+        # These items should be ordered in the same way as in the summary.
+        if self.bad_licenses:
+            recommendations.append(
+                _(
+                    "Fix bad licenses: At least one license in the LICENSES"
+                    " directory and/or provided by 'SPDX-License-Identifier'"
+                    " tags is invalid. They are either not valid SPDX License"
+                    " Identifiers or do not start with 'LicenseRef-'. FAQ about"
+                    " custom licenses:"
+                    " https://reuse.software/faq/#custom-license"
+                )
+            )
+        if self.deprecated_licenses:
+            recommendations.append(
+                _(
+                    "Fix deprecated licenses: At least one of the licenses in"
+                    " the LICENSES directory and/or provided by an"
+                    " 'SPDX-License-Identifier' tag or in '.reuse/dep5' has"
+                    " been deprecated by SPDX. The current list and their"
+                    " respective recommended  new identifiers can be found"
+                    " here: <https://spdx.org/licenses/#deprecated>"
+                )
+            )
+        if self.licenses_without_extension:
+            recommendations.append(
+                _(
+                    "Fix licenses without file extension: At least one license"
+                    " text file in the 'LICENSES' directory does not have a"
+                    " '.txt' file extension. Please rename the file(s)"
+                    " accordingly."
+                )
+            )
+        if self.missing_licenses:
+            recommendations.append(
+                _(
+                    "Fix missing licenses: For at least one of the license"
+                    " identifiers provided by the 'SPDX-License-Identifier'"
+                    " tags, there is no corresponding license text file in the"
+                    " 'LICENSES' directory. For SPDX license identifiers, you"
+                    " can simply run 'reuse download --all' to get any missing"
+                    " ones. For custom licenses (starting with 'LicenseRef-'),"
+                    " you need to add these files yourself."
+                )
+            )
+        if self.unused_licenses:
+            recommendations.append(
+                _(
+                    "Fix unused licenses: At least one of the license text"
+                    " files in 'LICENSES' is not referenced by any file, e.g."
+                    " by an 'SPDX-License-Identifier' tag. Please make sure"
+                    " that you either tag the accordingly licensed files"
+                    " properly, or delete the unused license text if you are"
+                    " sure that no file or code snippet is licensed as such."
+                )
+            )
+        if self.read_errors:
+            recommendations.append(
+                _(
+                    "Fix read errors: At least one of the files in your"
+                    " directory cannot be read by the tool. Please check the"
+                    " file permissions. You will find the affected files at the"
+                    " top of the output as part of the logged error messages."
+                )
+            )
+        if self.files_without_copyright or self.files_without_licenses:
+            recommendations.append(
+                _(
+                    "Fix missing copyright/licensing information: For one or"
+                    " more files, the tool cannot find copyright and/or"
+                    " licensing information. You typically do this by adding"
+                    " 'SPDX-FileCopyrightText' and 'SPDX-License-Identifier'"
+                    " tags to each file. The tutorial explains additional ways"
+                    " to do this: <https://reuse.software/tutorial/>"
+                )
+            )
+
+        return recommendations
+
+
+class ProjectSubsetReport:
+    """Like a :class:`ProjectReport`, but for a subset of the files using a
+    subset of features.
+    """
+
+    def __init__(self) -> None:
+        self.path: StrPath = ""
+        self.missing_licenses: dict[str, set[Path]] = {}
+        self.read_errors: set[Path] = set()
+        self.file_reports: set[FileReport] = set()
+
+    @classmethod
+    def generate(
+        cls,
+        project: Project,
+        subset_files: Collection[StrPath],
+        multiprocessing: bool = cpu_count() > 1,  # type: ignore
+    ) -> "ProjectSubsetReport":
+        """Generate a :class:`ProjectSubsetReport` from a :class:`Project`.
+
+        Args:
+            project: The :class:`Project` to lint.
+            subset_files: Only lint the files in this list.
+            multiprocessing: Whether to use multiprocessing.
+        """
+        subset_report = cls()
+        subset_report.path = project.root
+        results = _generate_file_reports(
+            project,
+            do_checksum=False,
+            subset_files=subset_files,
+            multiprocessing=multiprocessing,
+            add_license_concluded=False,
+        )
+        for result in results:
+            if result.error:
+                _process_error(result.error, result.path)
+                subset_report.read_errors.add(Path(result.path))
+                continue
+
+            file_report = cast(FileReport, result.report)
+            subset_report.file_reports.add(file_report)
+
+            for missing_license in file_report.missing_licenses:
+                subset_report.missing_licenses.setdefault(
+                    missing_license, set()
+                ).add(file_report.path)
+        return subset_report
+
+    @property
+    def files_without_licenses(self) -> set[Path]:
+        """Set of paths that have no licensing information."""
+        return {
+            file_report.path
+            for file_report in self.file_reports
+            if not file_report.licenses_in_file
+        }
+
+    @property
+    def files_without_copyright(self) -> set[Path]:
+        """Set of paths that have no copyright information."""
+        return {
+            file_report.path
+            for file_report in self.file_reports
+            if not file_report.copyright
+        }
+
+    @property
+    def is_compliant(self) -> bool:
+        """Whether the report subset is compliant with the REUSE Spec."""
+        return not any(
+            (
+                self.missing_licenses,
+                self.files_without_copyright,
+                self.files_without_licenses,
+                self.read_errors,
+            )
+        )
+
 
 class FileReport:  # pylint: disable=too-many-instance-attributes
     """Object that holds a linting report about a single file."""
@@ -388,32 +671,32 @@ class FileReport:  # pylint: disable=too-many-instance-attributes
         self.path = Path(path)
         self.do_checksum = do_checksum
 
-        self.reuse_infos: List[ReuseInfo] = []
+        self.reuse_infos: list[ReuseInfo] = []
 
         self.spdx_id: Optional[str] = None
         self.chk_sum: Optional[str] = None
-        self.licenses_in_file: List[str] = []
+        self.licenses_in_file: list[str] = []
         self.license_concluded: str = ""
         self.copyright: str = ""
 
-        self.bad_licenses: Set[str] = set()
-        self.missing_licenses: Set[str] = set()
+        self.bad_licenses: set[str] = set()
+        self.missing_licenses: set[str] = set()
 
-    def to_dict_lint(self) -> Dict[str, Any]:
+    def to_dict_lint(self) -> dict[str, Any]:
         """Turn the report into a json-like dictionary with exclusively
         information relevant for linting.
         """
         return {
-            # This gets rid of the './' prefix. In Python 3.9, use
-            # str.removeprefix.
             "path": PurePath(self.name).as_posix(),
             "copyrights": [
                 {
                     "value": line,
                     "source": reuse_info.source_path,
-                    "source_type": reuse_info.source_type.value
-                    if reuse_info.source_type
-                    else None,
+                    "source_type": (
+                        reuse_info.source_type.value
+                        if reuse_info.source_type
+                        else None
+                    ),
                 }
                 for reuse_info in self.reuse_infos
                 for line in reuse_info.copyright_lines
@@ -422,9 +705,11 @@ class FileReport:  # pylint: disable=too-many-instance-attributes
                 {
                     "value": str(expression),
                     "source": reuse_info.source_path,
-                    "source_type": reuse_info.source_type.value
-                    if reuse_info.source_type
-                    else None,
+                    "source_type": (
+                        reuse_info.source_type.value
+                        if reuse_info.source_type
+                        else None
+                    ),
                 }
                 for reuse_info in self.reuse_infos
                 for expression in reuse_info.spdx_expressions
@@ -467,8 +752,12 @@ class FileReport:  # pylint: disable=too-many-instance-attributes
                     # A license expression akin to Apache-1.0+ should register
                     # correctly if LICENSES/Apache-1.0.txt exists.
                     identifiers = {identifier}
-                    if identifier.endswith("+"):
-                        identifiers.add(identifier[:-1])
+                    if (
+                        plus_identifier := _strip_plus_from_identifier(
+                            identifier
+                        )
+                    ) != identifier:
+                        identifiers.add(plus_identifier)
                     # Bad license
                     if not identifiers.intersection(project.license_map):
                         report.bad_licenses.add(identifier)
